# -*- coding: utf-8 -*-
"""
Created on Thu Jul 27 08:53:36 2023

@author: pawelzik
"""
import numpy as np
from scipy.optimize import curve_fit
from PostPlot import PostPlot
import pandas as pd
from pathlib import Path

class reg_and_corr:

    def __init__(self, trace_subplot5 = '', legend_loc = 'upper left', \
                 legend_bbox_to_anchor = (1.09, 1)):
        
        self.postplot = PostPlot(trace_subplot5 = trace_subplot5)
        self.K_phases = None
        self.phase_t0 = None
        self.legend_loc = legend_loc 
        self.legend_bbox_to_anchor = legend_bbox_to_anchor


    
    def calc_correlation(self, func_1 , func_2 , state = 'False'):
        
        if state == True:
        
            # calulate correlation between two parameter from data frame
            corr_coeff= np.corrcoef(func_1, func_2)[0][1]
                
            annotate_string_corr_coeff = "$R_{\phi(S_{21}),Temp_{DUT}}$ = %.3f" % corr_coeff
        
            self.postplot.edit_annotation_in_plot(annotate_string_corr_coeff)
    
    def calc_regression_coeff_phase_S21(self, data_frame, time_unit = 'min', state = False):
            
        if state == True:
            phases = data_frame.S21_PHASE.tolist()
            self.phase_t0 = phases[0] 
    
            self.K_phases = np.median(data_frame.S21_PHASE)
            
            # time stamp of index = 0 is the start point of time axis, Michael
            # substract all other timestamps with time stamp of index zero to get time scale in
            # seconds, Michael
            
            time_sec = data_frame.TIMESTAMP - data_frame.TIMESTAMP[0]
            
            # set scaling of time axis depending on the time unit given in parameter list, Michael
            # default unit is minutes, Michael
            time_vals = self.scaling_time_axes(time_sec, time_unit)
            
            func, popt = self.choose_fit(time_vals, data_frame.S21_PHASE)
            
            annotate_string_reg_coeff = self.plot_fitted_func_param(func, *popt, time_unit = time_unit)
            
            self.postplot.edit_annotation_in_plot(annotate_string_reg_coeff)
        
    

    
    def plot_phase_curve_fit(self, data_frame, time_unit = 'min', state = False):
        
        if state == True:
            
            
            phases = data_frame.S21_PHASE.tolist()
            self.phase_t0 = phases[0] 

            self.K_phases = np.median(data_frame.S21_PHASE)
            
            # time stamp of index = 0 is the start point of time axis, Michael
            # substract all other timestamps with time stamp of index zero to get time scale in
            # seconds, Michael
            
            time_sec = data_frame.TIMESTAMP - data_frame.TIMESTAMP[0]
            
            # set scaling of time axis depending on the time unit given in parameter list, Michael
            # default unit is minutes, Michael
            time_vals = self.scaling_time_axes(time_sec, time_unit)
            
            # call chose fit function
            func, popt = self.choose_fit(time_vals, data_frame.S21_PHASE)
            
            # genrate trace for curvefit
            fit_phaeses = func(time_vals, *popt) 
            
            # determine max, min of measured phases
            min_phases, max_phases = self.get_extended_min_max(data_frame.S21_PHASE)
            
            # determine max, min of curvefit
            min_fit_phases, max_fit_phases = self.get_extended_min_max(fit_phaeses)
            
            # determine min of measure phases and min of curvefit 
            minimum = min(min_phases, min_fit_phases)
            
            # determine max of measured phases and max of curvefit
            maximum = max(max_phases, max_fit_phases)
            
            self.postplot.add_curvefit_to_plot(xvals =time_vals, \
                                               yvals = func(time_vals, *popt), \
                                               y_lim =[minimum, maximum],  \
                                               trace_color = 'green', \
                                               trace_label = 'fitted by\n' + func.__name__)
            
        else:
            self.postplot.add_curvefit_to_plot(xvals =[], \
                                   yvals = [], \
                                   y_lim = None,  \
                                   trace_color = 'None', \
                                   trace_label = '')  



# step response of PT1 with initial behavior , Michael    
    def PT1(self, x, K, T1):
        return K * (1 - np.exp(-x/T1)) + self.phase_t0 * np.exp(-x/T1)
    
    # step response of PT2 (D > 1) with initial behavior 
    def aperiodicPT2(self, x, K, T1, T2,):
        return K + self.phase_t0* np.exp(-x/T1)/(T2-T1) - K *T1* np.exp(-x/T1)/(T2-T1) \
                    - self.phase_t0 * np.exp(-x/T2)/(T2-T1) + K* T2* np.exp(-x/T2)/(T2-T1)
    
    # step response of PT2 (D = 1) with initial behavior, Michael
    def aper_borderPT2(self, x, K, T):
        return K + self.phase_t0*np.exp(-x/T) - K * np.exp(-x/T) + self.phase_t0*np.exp(-x/T)*x/T + \
            - K * np.exp(-x/T) * x/T

    # step response of PT2 (0 < D < 1) with initial behavior, Michael 
    def periodicPT2(self, x, D, T):
        return self.K_phases + (self.phase_t0 -self.K_phases)* \
            np.exp(-D*x/T)*np.cos(np.sqrt(1-np.square(D))*x/T) + \
            (self.phase_t0 - self.K_phases) * D * \
                np.exp(-D*x/T)*np.sin(np.sqrt(1-np.square(D))*x/T)/np.sqrt(1-np.square(D))
        
    # step response series of PT2 ( D = 1) and PT1 with initial behavior, Michael
    def aper_borderPT2_PT1(self, x, a, b, c, d, T1, T2):
        return a + b*np.exp(-x/T1)+c*x*np.exp(-x/T1) + d*np.exp(-x/T2)
    
    # method for cuvefit, Michael
    # two different algorithms for curvefit are use depending on used step response for fit, Michael
    def curvefit(self, func, time, phases):

        if func.__name__ == 'periodicPT2':
            popt, pcov = curve_fit(func, time, phases, method = 'trf', \
                                   bounds = ([0.001, 0.001], [0.99, np.inf]))
        else:    
            popt, pcov = curve_fit(func, time, phases, method = 'lm')

        return popt, pcov
    
    # method adds the parameter of the fit with the lowest error to annotation string, Michael

    def plot_fitted_func_param(self, func, *popt, time_unit):

        if func.__name__ == 'PT1':
            K = popt[0] 
            T1 = popt[1]
            annotate_string = "K: %.2f\n$T_1$: %.3f %s" %(K, T1, time_unit) 
        elif func.__name__ == 'aperiodicPT2':
            K = popt[0]
            T1 = popt[1]
            T2 = popt[2]
            annotate_string = "K: %.2f\n$T_1$: %.3f %s\n$T_2$: %.3f %s" % (K, T1, time_unit, T2, \
                                                                           time_unit)
        elif func.__name__ == 'aper_borderPT2':
            K = popt[0]
            T = popt[1]
            annotate_string = "K: %.2f\nT: %.3f %s" %(K, T, time_unit) 
        elif func.__name__ == 'periodicPT2':
            K = self.K_phases
            D = popt[0]
            T = popt[1]
            T_2perc = 4*T/D
            annotate_string = "K: %.2f\nD: %.3f\nT: %.3f %s\n$T_E$(2%%): %.2f %s" \
                % (K, D, T, time_unit, T_2perc, time_unit)
        elif func.__name__ == 'aper_borderPT2_PT1':
            T1 = popt[4]
            T2 = popt[5]
            annotate_string = "$T_1$: %.3f %s\n$T_2$: %.3f %s" % (T1, time_unit, T2, time_unit)
        else:
            # default value
            annotate_string =''
        return annotate_string
    
    # calculates the fit for all step responses and calulates the lowest error, Michael
    # parameter of the plot with the lowest error is returned by this function, Michael
    # if curve fit fails for one stept response it takes the next one and plots in command
    # window the functions
    # which failes, Michael
    def choose_fit(self, time, phases):
        popt = []
        perr = []
        used_functions = []
        functions = [self.PT1, self.aper_borderPT2, self.aperiodicPT2, self.periodicPT2, \
                     self.aper_borderPT2_PT1] 
        for func in functions:
            try:
                # call curvefit method and put the fitted parameter for step response function
                # to a list, Michael
                popt.append(self.curvefit(func, time, phases)[0])
                # calculation of fitting error
                perr.append(np.sqrt(np.square(np.subtract(phases, func(time, *popt[-1]))).mean()))
                used_functions.append(func)
            except RuntimeError:
                    print('curvefit of %s fails' %func.__name__)
        # get position of parameters in list which has the lowest calculated error, Michael
        pos = perr.index(min(perr))
        # return function name and parameter with the lowest calculated error during curvefit, Michael
        return used_functions[pos], popt[pos]

    

    # scaling time axes depending on the time unit
    def scaling_time_axes(self, time_sec, time_unit):
        if time_unit == 'min':
            time_vals = time_sec/60
        
        elif time_unit == 'hours':
            time_vals = time_sec/3600
        
        elif time_unit == 'sec':
            time_vals = time_sec
            
        else:
            time_vals = time_sec/60
    
        return time_vals
    
        # add 5 % of the distance between min and max to the range
    @staticmethod
    def get_extended_min_max(array):
        distance = array.max() - array.min()
        if distance == 0.:
            distance = 1
        return array.min()-0.05*distance, array.max()+0.05*distance


# for manually redo post plot after measurement has finished and insert analysis function, Michael
if __name__ == '__main__':
    
    
    # set result path for post plot 
    Results_Path = r'C:\git\climate-lab-test-stand\Python_script\TestData_JBY240'
    
    
    # set time unit for post post plot
    # default is minutes if entry in parameterlist left empty
    # possible entries: 'min' for minutes, 'hours' for hours and 'sec' for seconds
    time_unit = 'min'
    
    # set storepath for the post plots, Michael
    storepath = Results_Path + '\\AnalysisPlots' 
    
    # search all csv files in results folder 
    csv_file_list = list(Path(Results_Path).glob("**/*.csv"))
    
    # choose annotation option
    plot_correlation_coeff = False
    plot_regression_coeff = False
    
    # activate plot curvefitting
    
    plot_trace_curvefit = True
    
    # selction measurment data should be plotted in subplot5
    # 'logger sens' : values for temperature and humidity of logger sensor in
    # measurement instrument chmaber
    # 'chamber_sens' : values for temepratere and humidity readback from chamber sensors
    # of chamber for measurement instruments
    # 'heater_dut_chamber': activity of temp heater and hum heater readback from DUT chamber
    # This is also the default parameter
    
    trace_selection = ""
    
    
    # create postplot object, Michael
    analysisplot_obj = reg_and_corr(trace_subplot5= trace_selection)
        
    
    # empty data frame for concat the data frames from csv import to plot full transistion
    concat_data_frame = pd.DataFrame()
    
    # plot results for each csv-file
    for index, csv_file in enumerate(csv_file_list):
    
        # import csv-data from csv-files in list, Michael
        data_frame = analysisplot_obj.postplot.import_csv(str(csv_file))
        
        # determine title of plot from csv-filename, Michael
        title = csv_file.name
        
        # concate datesframe for plotting full transistion data 
        concat_data_frame = pd.concat([concat_data_frame,data_frame],ignore_index=True, sort = False)       
        
        
        # plot frame data
        analysisplot_obj.postplot.plot_frame_data(data_frame, title, time_unit) 
        
            
        # determine correlation coefficient between func 1 and func 2 and plot coefficient
        analysisplot_obj.calc_correlation(func_1 = data_frame.S21_PHASE , \
                                  func_2 = data_frame.TEMP_DUT, state = plot_correlation_coeff )
       
     
        
            
        # start curvefit and determine coefficients best fit function and plot coefficients
        analysisplot_obj.calc_regression_coeff_phase_S21(data_frame, time_unit = 'min', \
                                                      state = plot_regression_coeff)
        
            
        analysisplot_obj.plot_phase_curve_fit(data_frame = data_frame, time_unit = time_unit, \
                                          state = plot_trace_curvefit)
        
        # set filename of post plot, Michael
        filename = str(csv_file.stem) + '.pdf'
        
        # store post plot under the path taht is set in storepath with the earlier defined filename
        analysisplot_obj.postplot.save_fig(storepath, filename)
    
      
    # plot of all steps of a sweep is plotted in one diagram, Michael
    # title of this plot will be always full transistion, Michel
    analysisplot_obj.postplot.plot_frame_data(concat_data_frame, 'Full Transistion', time_unit)
    
    analysisplot_obj.plot_phase_curve_fit(data_frame = concat_data_frame, time_unit = time_unit, \
                                      state = False)
    
    # filename of plot that contains all steps of sweep is set to FullTransistion, Michael
    filename = 'Full_Transistion' + '.pdf'
    
    # plot with the results of all steps is store under the predefined storpath with the 
    # earlier defined filename, Michael
    analysisplot_obj.postplot.save_fig(storepath, filename)    
  